# Copyright (c) 2024-present yowoda
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Updated automatically by the deploy workflow
__version__ = "1.1.2"
